#!/usr/bin/env python3
'''Various tools to check physical properties.'''
import numpy as np
from scipy.optimize import minimize_scalar

from .dft import get_epsilon
from .logger import log


def cutoff2gridspacing(E):
    '''Convert plane wave energy cut-off to a real-space grid spacing.

    Reference: Phys. Rev. B 54, 14362.

    Args:
        E (float): Energy in Hartree.

    Returns:
        float: Grid spacing in Bohr.
    '''
    return np.pi / np.sqrt(2 * E)


def gridspacing2cutoff(h):
    '''Convert real-space grid spacing to plane wave energy cut-off.

    Reference: Phys. Rev. B 54, 14362.

    Args:
        h (float): Grid spacing in Bohr.

    Returns:
        float: Cut-off in Hartree.
    '''
    return 0.5 * (np.pi / h)**2


def center_of_mass(coords, masses=None):
    '''Calculate the center of mass for a set of coordinates and masses.

    Args:
        coords (ndarray): Array of real-space coordinates.

    Keyword Args:
        masses (ndarray): Mass or weight for each coordinate.

    Returns:
        ndarray: Center of mass.
    '''
    if masses is None:
        masses = np.ones(len(coords))

    return np.sum(masses * coords.T, axis=1) / np.sum(masses)


def inertia_tensor(coords, masses=None):
    '''Calculate the inertia tensor for a set of coordinates and masses.

    Args:
        coords (ndarray): Array of real-space coordinates.

    Keyword Args:
        masses (ndarray): Mass or weight for each coordinate.

    Returns:
        ndarray: Inertia tensor.
    '''
    if masses is None:
        masses = np.ones(len(coords))

    # The inertia tensor for a set of point masses can be calculated with simple summation
    # https://en.wikipedia.org/wiki/Moment_of_inertia#Definition_2
    I = np.empty((3, 3))
    I[0, 0] = np.sum(masses * (coords[:, 1]**2 + coords[:, 2]**2))
    I[1, 1] = np.sum(masses * (coords[:, 0]**2 + coords[:, 2]**2))
    I[2, 2] = np.sum(masses * (coords[:, 0]**2 + coords[:, 1]**2))

    I[0, 1] = I[1, 0] = -np.sum(masses * (coords[:, 0] * coords[:, 1]))
    I[0, 2] = I[2, 0] = -np.sum(masses * (coords[:, 0] * coords[:, 2]))
    I[1, 2] = I[2, 1] = -np.sum(masses * (coords[:, 1] * coords[:, 2]))
    return I


def get_dipole(scf, n=None):
    '''Calculate the electric dipole moment.

    Reference: J. Chem. Phys. 155, 224109.

    Args:
        scf: SCF object.

    Keyword Args:
        n (float): Real-space electronic density.

    Returns:
        ndarray: Electric dipole moment in e * Bohr.
    '''
    # The dipole may be extremely large. This can be because of periodic boundary conditions,
    # e.g., the density gets "smeared" to the edges if the atom sits at one edge.
    # One fix can be to center the atom/molecule inside the box.
    atoms = scf.atoms
    if n is None:
        n = scf.n
    if scf.n is None:
        log.error('There is no density to calculate a dipole.')
        return 0

    # Diple moment: mu = \sum Z X - \int n(r) r dr
    mu = np.array([0, 0, 0], dtype=float)
    for i in range(atoms.Natoms):
        mu += atoms.Z[i] * atoms.X[i]

    dV = atoms.Omega / np.prod(atoms.s)
    for dim in range(3):
        mu[dim] -= dV * np.sum(n * atoms.r[:, dim])
    return mu


def get_ip(scf):
    '''Calculate the ionization potential by calculating the negative HOMO energy.

    Reference: Physica 1, 104.

    Args:
        scf: SCF object.

    Returns:
        float: Ionization potential in Hartree.
    '''
    epsilon = get_epsilon(scf, scf.W)
    # Add up spin states
    epsilon = np.sum(epsilon, axis=0)
    return -epsilon[-1]


def check_ortho(object, func, eps=1e-9):
    '''Check the orthogonality condition for a set of functions.

    Args:
        object: Atoms or SCF object.
        func (ndarray): Discretized set of functions.

    Keyword Args:
        eps (float): Tolerance for the condition.

    Returns:
        bool: Orthogonality status for the set of functions.
    '''
    func = np.atleast_3d(func)
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    # It makes no sense to calculate anything for only one function
    if atoms.Nstate == 1:
        log.warning('Need at least two functions to check their orthogonality.')
        return True

    # We integrate over our cell, the integration borders then become a=0 and b=cell length
    # The integration prefactor dV is (b-a)/n, with n as the sampling
    # For a 3d integral we have to multiply for every direction
    dV = atoms.Omega / np.prod(atoms.s)

    ortho_bool = True
    # Check the condition for every combination
    # Orthogonality condition: \int func1^* func2 dr = 0
    for spin in range(atoms.Nspin):
        for i in range(atoms.Nstate):
            for j in range(i + 1, atoms.Nstate):
                res = dV * np.sum(func[spin, :, i].conj() * func[spin, :, j])
                tmp_bool = abs(res) < eps
                ortho_bool *= tmp_bool
                log.debug(f'Function {i} and {j}:\n\tValue: {res:.7f}\n\tOrthogonal: {tmp_bool}')
    log.info(f'Orthogonal: {ortho_bool}')
    return ortho_bool


def check_norm(object, func, eps=1e-9):
    '''Check the normalization condition for a set of functions.

    Args:
        object: Atoms or SCF object.
        func (ndarray): Discretized set of functions.

    Keyword Args:
        eps (float): Tolerance for the condition.

    Returns:
        bool: Normalization status for the set of functions.
    '''
    func = np.atleast_3d(func)
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    # We integrate over our cell, the integration borders then become a=0 and b=cell length
    # The integration prefactor dV is (b-a)/n, with n as the sampling
    # For a 3d integral we have to multiply for every direction
    dV = atoms.Omega / np.prod(atoms.s)

    norm_bool = True
    # Check the condition for every function
    # Normality condition: \int func^* func dr = 1
    for spin in range(atoms.Nspin):
        for i in range(atoms.Nstate):
            res = dV * np.sum(func[spin, :, i].conj() * func[spin, :, i])
            tmp_bool = abs(1 - res) < eps
            norm_bool *= tmp_bool
            log.debug(f'Function {i}:\n\tValue: {res:.7f}\n\tNormalized: {tmp_bool}')
    log.info(f'Normalized: {norm_bool}')
    return norm_bool


def check_orthonorm(object, func):
    '''Check the orthonormality conditions for a set of functions.

    Args:
        object: Atoms or SCF object.
        func (ndarray): Discretized set of functions.

    Returns:
        bool: Orthonormality status for the set of functions.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    ortho_bool = check_ortho(atoms, func)
    norm_bool = check_norm(atoms, func)
    log.info(f'Orthonormal: {ortho_bool * norm_bool}')
    return ortho_bool * norm_bool


def get_isovalue(n, percent=85):
    '''Find an isovalue that contains a specified percentage of the electronic density n.

    Args:
        n (float): Real-space electronic density.

    Keyword Args:
        percent (float): Amount of density that should be contained.

    Returns:
        float: Isovalue that contains the specified percentage of the density.
    '''
    def deviation(isovalue):
        n_mask = np.sum(n[n > isovalue])
        return abs(percent - (n_mask / n_ref) * 100)

    # Integrated density
    n_ref = np.sum(n)
    # Finding the isovalue is an optimization problem, minimizing the deviation above
    # The problem is bound by zero (no density) and the maximum value in n
    res = minimize_scalar(deviation, bounds=(0, np.max(n)), method='bounded')
    return res.x



#!/usr/bin/env python3
'''Extra functions that need additional dependencies to work.

To also install all additional dependencies, use::

    pip install eminus[all]

Alternativle, you can only install selected extras using the respective name:

* :mod:`~eminus.extras.fods`
* :mod:`~eminus.extras.libxc`
* :mod:`~eminus.extras.viewer`
'''
from .fods import get_fods, remove_core_fods, split_fods
from .libxc import libxc_functional
from .viewer import view, view_atoms, view_file

__all__ = ['get_fods', 'libxc_functional', 'remove_core_fods', 'split_fods', 'view', 'view_atoms',
           'view_file']



#!/usr/bin/env python3
'''SCF class definition.'''
import copy
import logging
import time

import numpy as np

from .dft import guess_gaussian, guess_pseudo, guess_random
from .energies import Energy, get_Eewald, get_Esic
from .gth import init_gth_loc, init_gth_nonloc
from .io import read_gth
from .logger import create_logger, get_level
from .minimizer import cg, lm, pccg, pclm, sd  # noqa: F401
from .potentials import init_pot
from .version import info
from .xc import XC_MAP


class SCF:
    '''Perform direct minimizations.

    Args:
        atoms: Atoms object.

    Keyword Args:
        xc (str): Comma-separated exchange-correlation functional description (case insensitive).

            Adding 'libxc:' before a functional will try to use the LibXC interface.

            Example: 'lda,pw'; 'lda,'; ',vwn'; ','; 'libxc:LDA_X,libxc:7',
            Default: 'lda,vwn'
        pot (str): Type of pseudopotential (case insensitive).

            Example: 'GTH'; 'harmonic'; 'Coulomb'; 'Ge',
            Default: 'gth'
        guess (str): Initial guess method for the basis functions (case insensitive).

            Example: 'Gauss'; 'gaussian'; 'random'; 'rand',
            Default: 'gaussian'
        etol (float): Convergence tolerance of the total energy.

            Default: 1e-7
        cgform (int): Conjugated-gradient form for the pccg minimization.

            1 for Fletcher-Reeves, 2 for Polak-Ribiere, and 3 for Hestenes-Stiefel.

            Default: 1
        min (dict | None): Dictionary to set the order and number of steps per minimization method.

            Example: {'sd': 10, 'pccg': 100}; {'pccg': 10, 'lm': 25, 'pclm': 50},
            Default: None (will default to {'pccg': 250})
        sic (bool): Calculate the Kohn-Sham Perdew-Zunger SIC energy at the end of the SCF step.

            Default: False
        verbose (int | str): Level of output (case insensitive).

            Can be one of 'CRITICAL', 'ERROR', 'WARNING', 'INFO', or 'DEBUG'.
            An integer value can be used as well, where larger numbers mean more output, starting
            from 0.

            Default: 'info'
    '''
    def __init__(self, atoms, xc='lda,vwn', pot='gth', guess='gaussian', etol=1e-7, cgform=1,
                 sic=False, min=None, verbose=None):
        self.atoms = copy.copy(atoms)  # Atoms object
        self.xc = xc.lower()           # Exchange-correlation functional
        self.pot = pot.lower()         # Used pseudopotential
        self.guess = guess             # Initial wave functions guess
        self.etol = etol               # Total energy convergence tolerance
        self.cgform = cgform           # Conjugate gradient form
        self.sic = sic                 # Calculate the sic energy
        self.min = min                 # Minimization methods

        # Set min here, better not use mutable data types in signatures
        if self.min is None:
            # For systems with bad convergence throw in some sd steps
            self.min = {'sd': 25, 'pccg': 250}

        # Initialize logger
        self.log = create_logger(self)
        if verbose is None:
            self.verbose = atoms.verbose
        else:
            self.verbose = verbose

        # Set up final and intermediate results
        self.W = None             # Basis functions
        self.n = None             # Electronic density
        self.energies = Energy()  # Energy object that holds energy contributions
        self.clear()

        # Parameters that will be built out of the inputs
        self.GTH = {}             # Dictionary of GTH parameters per atom species
        self.Vloc = None          # Local pseudopotential contribution
        self.NbetaNL = 0          # Number of projector functions for the non-local gth potential
        self.prj2beta = None      # Index matrix to map to the correct projector function
        self.betaNL = None        # Atomic-centered projector functions
        self.print_precision = 6  # Precision of the energy in the minimizer logger
        self.initialize()

    def clear(self):
        '''Initialize and clear intermediate results.'''
        self.Y = None       # Orthogonal wave functions
        self.n_spin = None  # Electronic densities per spin
        self.phi = None     # Hartree field
        self.exc = None     # Exchange-correlation energy density
        self.vxc = None     # Exchange-correlation potential
        return self

    def initialize(self):
        '''Validate inputs, update them and build all necessary parameters.'''
        if not self.atoms.is_built:
            self.atoms.build()
        self._set_potential()
        self._init_W()
        self.print_precision = int(abs(np.log10(self.etol))) + 1
        return self

    def run(self, **kwargs):
        '''Run the self-consistent field (SCF) calculation.'''
        if self.log.level <= logging.DEBUG:
            info()
        self.log.debug(f'--- System information ---\n{self.atoms}\n'
                       f'Spin handling: {"un" if self.atoms.Nspin == 1 else ""}polarized\n'
                       f'Number of states: {self.atoms.Nstate}\n'
                       f'Occupation per state:\n{self.atoms.f}\n'
                       f'\n--- Cell information ---\nSide lengths: {self.atoms.a} Bohr\n'
                       f'Sampling per axis: {self.atoms.s}\n'
                       f'Cut-off energy: {self.atoms.ecut} Hartree\n'
                       f'Compression: {len(self.atoms.G2) / len(self.atoms.G2c):.5f}\n'
                       f'\n--- Calculation information ---\n{self}\n\n--- SCF data ---')

        # Calculate Ewald energy that only depends on the system geometry
        self.energies.Eewald = get_Eewald(self.atoms)

        exch, corr = self.xc.split(',')
        try:
            if 'libxc' not in exch:
                XC_MAP[exch]
        except KeyError:
            self.log.warning('Use a mock functional for the exchange part.')
        try:
            if 'libxc' not in corr:
                XC_MAP[corr]
        except KeyError:
            self.log.warning('Use a mock functional for the correlation part.')

        # Start minimization procedures
        Etots = []
        minimizer_log = {}
        for imin in self.min:
            try:
                self.log.info(f'Start {eval(imin).__name__}...')
            except NameError:
                self.log.exception(f'No minimizer found for "{imin}"')
                raise
            start = time.perf_counter()
            Elist = eval(imin)(self, self.min[imin], **kwargs)  # Call minimizer
            end = time.perf_counter()
            minimizer_log[imin] = {}  # Create an entry for the current minimizer
            minimizer_log[imin]['time'] = end - start  # Save time in dictionary
            minimizer_log[imin]['iter'] = len(Elist)  # Save iterations in dictionary
            Etots += Elist  # Append energies from minimizer
            # Do not start other minimizations if one converged
            if len(Etots) > 1 and abs(Etots[-2] - Etots[-1]) < self.etol:
                break
        if len(Etots) > 1 and abs(Etots[-2] - Etots[-1]) < self.etol:
            self.log.info(f'SCF converged after {len(Etots)} iterations.')
        else:
            self.log.warning('SCF not converged!')

        # Print SCF data
        self.log.debug('\n--- SCF results ---')
        t_tot = 0
        for imin in minimizer_log:
            N = minimizer_log[imin]['iter']
            t = minimizer_log[imin]['time']
            t_tot += t
            self.log.debug(f'Minimizer: {imin}'
                           f'\nIterations:\t{N}'
                           f'\nTime:\t\t{t:.5f} s'
                           f'\nTime/Iteration:\t{t / N:.5f} s')
        self.log.info(f'Total SCF time: {t_tot:.5f} s')

        # Calculate SIC energy if desired
        if self.sic:
            self.energies.Esic = get_Esic(self, self.Y)

        # Print energy data
        if self.log.level <= logging.DEBUG:
            self.log.debug(f'\n--- Energy data ---\n{self.energies}')
        else:
            self.log.info(f'Total energy: {self.energies.Etot:.9f} Eh')
        return self.energies.Etot

    kernel = run

    def _set_potential(self):
        '''Build the potential.'''
        atoms = self.atoms
        if self.pot == 'gth':
            for ia in range(atoms.Natoms):
                self.GTH[atoms.atom[ia]] = read_gth(atoms.atom[ia], atoms.Z[ia])
            # Set up the local and non-local part
            self.Vloc = init_gth_loc(self)
            self.NbetaNL, self.prj2beta, self.betaNL = init_gth_nonloc(self)
        else:
            self.Vloc = init_pot(self)
        return

    def _init_W(self):
        '''Initialize wave functions.'''
        if self.guess in ('gauss', 'gaussian'):
            # Start with gaussians at atom positions
            self.W = guess_gaussian(self, complex=True)
        elif self.guess in ('rand', 'random'):
            # Start with randomized, complex basis functions with a random seed
            self.W = guess_random(self, complex=True)
        elif self.guess in ('pseudo'):
            # Start with pseudo-random numbers, mostly to compare with SimpleDFT
            self.W = guess_pseudo(self, seed=1234)
        else:
            self.log.error(f'No guess found for "{self.guess}"')
        return

    def __repr__(self):
        '''Print the parameters stored in the SCF object.'''
        return f'XC functionals: {self.xc}\n' \
               f'Potential: {self.pot}\n' \
               f'Starting guess: {self.guess}\n' \
               f'Convergence tolerance: {self.etol}\n' \
               f'Non-local contribution: {self.NbetaNL > 0}'

    @property
    def verbose(self):
        '''Verbosity level.'''
        return self._verbose

    @verbose.setter
    def verbose(self, level):
        '''Verbosity setter to sync the logger with the property.'''
        self._verbose = get_level(level)
        self.log.setLevel(self._verbose)
        return


class RSCF(SCF):
    '''SCF class for spin-paired systems. '''  # noqa: D210
    __doc__ += SCF.__doc__

    def initialize(self):
        '''Validate inputs, update them and build all necessary parameters.'''
        super().initialize()
        self.atoms._set_states(Nspin=1)
        return self


class USCF(SCF):
    '''SCF class for spin-polarized systems. '''  # noqa: D210
    __doc__ += SCF.__doc__

    def initialize(self):
        '''Validate inputs, update them and build all necessary parameters.'''
        super().initialize()
        self.atoms._set_states(Nspin=2)
        return self



#!/usr/bin/env python3
'''Viewer functions for Jupyter notebooks.'''
import numpy as np

from .fods import split_fods
from ..data import COVALENT_RADII, CPK_COLORS
from ..io import create_pdb_str, read_cube, read_xyz
from ..logger import log
from ..tools import get_isovalue
from ..units import bohr2ang


def view(*args, **kwargs):
    '''Unified display function.'''
    if isinstance(args[0], str):
        return view_file(*args, **kwargs)
    else:
        return view_atoms(*args, **kwargs)


def view_atoms(object, extra=None, unit='bohr', plot_n=False, percent=85, surfaces=20):
    '''Display atoms and 3D-coordinates, e.g., FODs or grid points, or even densities.

    Args:
        object: Atoms or SCF object.

    Keyword Args:
        extra (ndarray | list ): Extra coordinates or FODs to display.
        unit (str): Units to display, can be either 'Bohr' (default) or 'Ang' (case insensitive).
        plot_n (bool): Weather to plot the electronic density (only for executed scf objects).
        percent (float): Amount of density that should be contained.
        surfaces (int): Number of surfaces to display in density plots (reduce for performance).

    Returns:
        None.
    '''
    try:
        import plotly.graph_objects as go
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    # Handle units
    if 'ang' in unit.lower():
        unit_str = 'Å'
        X = bohr2ang(atoms.X)
        a = bohr2ang(atoms.a)
        # FODs are saved in a list
        if isinstance(extra, list):
            extra = [bohr2ang(i) for i in extra]
        else:
            extra = bohr2ang(extra)
    else:
        unit_str = 'a<sub>0</sub>'
        X = atoms.X
        a = atoms.a

    fig = go.Figure()
    # Add species one by one to be able to have them named and be selectable in the legend
    # Note: The size scaling is mostly arbitray and has no meaning
    for ia in sorted(set(atoms.atom)):
        mask = np.where(np.asarray(atoms.atom) == ia)[0]
        atom_data = go.Scatter3d(x=X[mask, 0], y=X[mask, 1], z=X[mask, 2],
                                 name=ia,
                                 mode='markers',
                                 marker=dict(size=2 * np.pi * np.sqrt(COVALENT_RADII[ia]),
                                             color=CPK_COLORS[ia],
                                             line={'color': 'black', 'width': 2}))
        fig.add_trace(atom_data)
    if extra is not None:
        # If a list has been provided with the length of two it has to be FODs
        if isinstance(extra, list) and len(extra) == 2:
            if extra[0].size != 0:
                extra_data = go.Scatter3d(x=extra[0][:, 0], y=extra[0][:, 1], z=extra[0][:, 2],
                                          name='up-FOD',
                                          mode='markers',
                                          marker=dict(size=np.pi, color='red'))
                fig.add_trace(extra_data)
            if extra[1].size != 0:
                extra_data = go.Scatter3d(x=extra[1][:, 0], y=extra[1][:, 1], z=extra[1][:, 2],
                                          name='down-FOD',
                                          mode='markers',
                                          marker=dict(size=np.pi, color='green'))
                fig.add_trace(extra_data)
        # Treat extra as normal coordinates otherwise
        else:
            extra_data = go.Scatter3d(x=extra[:, 0], y=extra[:, 1], z=extra[:, 2],
                                      name='Coordinates',
                                      mode='markers',
                                      marker=dict(size=1, color='red'))
            fig.add_trace(extra_data)

    # A density can be plotted for an scf object
    if plot_n:
        try:
            if 'ang' in unit.lower():
                r = bohr2ang(atoms.r)
            else:
                r = atoms.r
            den_data = go.Volume(x=r[:, 0], y=r[:, 1], z=r[:, 2], value=object.n,
                                 name='Density',
                                 colorbar_title=f'Density ({percent}%)',
                                 colorscale='Inferno',
                                 isomin=get_isovalue(object.n, percent=percent),
                                 isomax=np.max(object.n),
                                 surface_count=surfaces,
                                 opacity=0.1,
                                 showlegend=True)
            fig.add_trace(den_data)
            # Move colorbar to the left
            fig.data[-1].colorbar.x = -0.15
        except (AttributeError, TypeError):
            log.warning('Density plots are only possible for executed SCF objects.')

    # Theming
    fig.update_layout(scene={'xaxis': {'range': [0, a[0]], 'title': f'x [{unit_str}]'},
                             'yaxis': {'range': [0, a[1]], 'title': f'y [{unit_str}]'},
                             'zaxis': {'range': [0, a[2]], 'title': f'z [{unit_str}]'},
                             'aspectmode': 'cube'},
                      legend={'itemsizing': 'constant', 'title': 'Selection'},
                      hoverlabel_bgcolor='black',
                      template='none')
    fig.show()
    return


def view_file(filename, isovalue=0.01, gui=False, elec_symbols=None, **kwargs):
    '''Display molecules and orbitals.

    Reference: Bioinformatics 34, 1241.

    Args:
        filename (str): Input file path/name. This can be either a cube or xyz file.

    Keyword Args:
        isovalue (float): Isovalue for sizing orbitals.
        gui (bool): Turn on the NGLView GUI.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        NGLWidget: Viewable object.
    '''
    try:
        from nglview import NGLWidget, TextStructure
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[viewer]".\n\n')
        raise

    if elec_symbols is None:
        elec_symbols = ['X', 'He']

    if isinstance(isovalue, str):
        isovalue = float(isovalue)
    view = NGLWidget(**kwargs)
    view._set_size('400px', '400px')

    if filename.endswith('.xyz'):
        # Atoms
        atom, X = read_xyz(filename)
        atom, X, X_fod = split_fods(atom, X, elec_symbols)
        view.add_component(TextStructure(create_pdb_str(atom, X)))
        view[0].clear()
        view[0].add_ball_and_stick()
        # Spin up FODs
        if len(X_fod[0]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(X_fod[0]),
                               X_fod[0])))
            view[1].clear()
            view[1].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
        # Spin down FODs
        if len(X_fod[1]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(X_fod[1]),
                               X_fod[1])))
            view[2].clear()
            view[2].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
        view.center()

    if filename.endswith('.cube'):
        # Atoms and cell
        atom, X, _, a, _ = read_cube(filename)
        atom, X, X_fod = split_fods(atom, X, elec_symbols)
        view.add_component(TextStructure(create_pdb_str(atom, X, a)))
        view[0].clear()
        view[0].add_ball_and_stick()
        view.add_unitcell()
        view.center()
        # Spin up
        view.add_component(filename)
        view[1].clear()
        # Negate isovalue here as a workaround for some display bugs
        view[1].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=-isovalue,
                            color='lightgreen',
                            opacity=0.75,
                            side='front')
        # Spin down
        view.add_component(filename)
        view[2].clear()
        view[2].add_surface(negateIsolevel=True,
                            isolevelType='value',
                            isolevel=isovalue,
                            color='red',
                            opacity=0.75,
                            side='front')
        # Spin up FODs
        if len(X_fod[0]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[0]] * len(X_fod[0]),
                               X_fod[0])))
            view[3].clear()
            view[3].add_ball_and_stick(f'_{elec_symbols[0]}', color='red', radius=0.1)
        # Spin down FODs
        if len(X_fod[1]) > 0:
            view.add_component(TextStructure(create_pdb_str([elec_symbols[1]] * len(X_fod[1]),
                               X_fod[1])))
            view[4].clear()
            view[4].add_ball_and_stick(f'_{elec_symbols[1]}', color='green', radius=0.1)
    return view.display(gui=gui)



#!/usr/bin/env python3
'''Utilities to localize and analyze orbitals.'''
import numpy as np
from scipy.linalg import eig, norm

from .logger import log
from .utils import handle_spin_gracefully


def eval_psi(atoms, psi, r):
    '''Evaluate orbitals at given coordinate points.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        r (ndarray): Real-space positions.

    Returns:
        ndarray: Values of psi at points r.
    '''
    # Shift the evaluation point to (0,0,0), because we always have a lattice point there
    psi_T = atoms.T(psi, -r)
    psi_Trs = atoms.I(psi_T)
    # The zero entry is always the value at point (0,0,0)
    return psi_Trs[0]


def get_R(atoms, psi, fods):
    '''Calculate transformation matrix to build Fermi orbitals from Kohn-Sham orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (ndarray): Fermi-orbital descriptors.

    Returns:
        ndarray: Transformation matrix R.
    '''
    # We only calculate occupied orbitals, so a zero matrix is enough
    R = np.empty((len(fods), len(fods)), dtype=complex)

    for i in range(len(fods)):
        # Get the value at one FOD position for all psi
        psi_fod = eval_psi(atoms, psi, fods[i])
        sum_psi_fod = np.sqrt(np.sum(psi_fod.conj() * psi_fod))
        for j in range(len(fods)):
            R[i, j] = psi_fod[j].conj() / sum_psi_fod
    return R


def get_FO(atoms, psi, fods):
    '''Calculate Fermi orbitals from Kohn-Sham orbitals and a set of respective FODs.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (ndarray): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi orbitals.
    '''
    FO = np.zeros((atoms.Nspin, len(atoms.r), atoms.Nstate), dtype=complex)

    # Transform psi to real-space
    psi_rs = atoms.I(psi)
    for spin in range(atoms.Nspin):
        # Get the transformation matrix R
        R = get_R(atoms, psi[spin], fods[spin])
        for i in range(len(R)):
            for j in range(atoms.Nstate):
                FO[spin, :, i] += R[i, j] * psi_rs[spin, :, j]
    return FO


def get_S(atoms, psirs):
    '''Calculate overlap matrix between orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Overlap matrix S.
    '''
    # Overlap elements: S_ij = \int psi_i^* psi_j dr
    S = np.empty((atoms.Nstate, atoms.Nstate), dtype=complex)

    dV = atoms.Omega / np.prod(atoms.s)
    for i in range(atoms.Nstate):
        for j in range(atoms.Nstate):
            S[i, j] = dV * np.sum(psirs[:, i].conj() * psirs[:, j])
    return S


def get_FLO(atoms, psi, fods):
    '''Calculate Fermi-Löwdin orbitals by orthonormalizing Fermi orbitals.

    Reference: J. Chem. Phys. 153, 084104.

    Args:
        atoms: Atoms object.
        psi (ndarray): Set of orbitals in reciprocal space.
        fods (ndarray): Fermi-orbital descriptors.

    Returns:
        ndarray: Real-space Fermi-Löwdin orbitals.
    '''
    FO = get_FO(atoms, psi, fods)
    FLO = np.empty((atoms.Nspin, len(atoms.r), atoms.Nstate), dtype=complex)

    for spin in range(atoms.Nspin):
        # Calculate the overlap matrix for FOs
        S = get_S(atoms, FO[spin])
        # Calculate eigenvalues and eigenvectors
        Q, T = eig(S)
        # Löwdins symmetric orthonormalization method
        Q12 = np.diag(1 / np.sqrt(Q))
        FLO[spin] = FO[spin] @ (T @ Q12 @ T.T)
    return FLO


@handle_spin_gracefully
def wannier_cost(atoms, psirs):
    '''Calculate the Wannier cost function, namely the orbital variance. Equivalent to Foster-Boys.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Variance per orbital.
    '''
    # Variance = \int psi r^2 psi - (\int psi r psi)^2
    centers = wannier_center(atoms, psirs)
    moments = second_moment(atoms, psirs)
    costs = moments - norm(centers, axis=1)**2
    log.debug(f'Centers:\n{centers}\nMoments:\n{moments}')
    log.info(f'Costs:\n{costs}')
    return costs


def wannier_center(atoms, psirs):
    '''Calculate Wannier centers, i.e., the expectation values of r.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Wannier centers per orbital.
    '''
    dV = atoms.Omega / np.prod(atoms.s)

    centers = np.empty((atoms.Nstate, 3))
    for i in range(atoms.Nstate):
        for dim in range(3):
            centers[i, dim] = dV * np.real(np.sum(psirs[:, i].conj() * atoms.r[:, dim] *
                                           psirs[:, i], axis=0))
    return centers


def second_moment(atoms, psirs):
    '''Calculate the second moments, i.e., the expectation values of r^2.

    Reference: J. Chem. Phys. 137, 224114.

    Args:
        atoms: Atoms object.
        psirs (ndarray): Set of orbitals in real-space.

    Returns:
        ndarray: Second moments per orbital.
    '''
    dV = atoms.Omega / np.prod(atoms.s)
    r2 = norm(atoms.r, axis=1)**2

    moments = np.empty(atoms.Nstate)
    for i in range(atoms.Nstate):
        moments[i] = dV * np.real(np.sum(psirs[:, i].conj() * r2 * psirs[:, i], axis=0))
    return moments



#!/usr/bin/env python3
'''Basis set dependent operators for a plane wave basis.

These operators act on discretized wave functions, i.e., the arrays W.

These W are column vectors. This has been chosen to let theory and code coincide, e.g.,
W^dagger W becomes W.conj().T @ W.

The downside is that the i-th state will be accessed with W[:, i] instead of W[i].
Choosing the i-th state makes the array 1d.

These operators can act on six different options, namely

1. the real-space
2. the real-space (1d)
3. the full reciprocal space
4. the full reciprocal space (1d)
5. the active reciprocal space
6. the active reciprocal space (1d)

The active space is the truncated reciprocal space by restricting it with a sphere given by ecut.

Every spin dependence will be handled with handle_spin_gracefully by calling the operators for each
spin individually.
'''
import os

import numpy as np
from scipy.fft import fftn, ifftn

from .utils import handle_spin_gracefully

try:
    THREADS = int(os.environ['OMP_NUM_THREADS'])
except KeyError:
    THREADS = None


# Spin handling is trivial for this operator
def O(atoms, W):
    '''Overlap operator.

    This operator acts on the options 3, 4, 5, and 6.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    return atoms.Omega * W


@handle_spin_gracefully
def L(atoms, W):
    '''Laplacian operator.

    This operator acts on the options 3 and 5.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    # G2 is a normal 1d row vector, reshape it so it can be applied to the column vector W
    if len(W) == len(atoms.G2c):
        G2 = atoms.G2c[:, None]
    else:
        G2 = atoms.G2[:, None]
    return -atoms.Omega * G2 * W


@handle_spin_gracefully
def Linv(atoms, W):
    '''Inverse Laplacian operator.

    This operator acts on the options 3 and 4.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    out = np.empty_like(W, dtype=complex)
    # Ignore the division by zero for the first elements
    with np.errstate(divide='ignore', invalid='ignore'):
        if W.ndim == 1:
            # One could do some proper indexing with [1:] but indexing is slow
            out = W / atoms.G2 / -atoms.Omega
            out[0] = 0
        else:
            # G2 is a normal 1d row vector, reshape it so it can be applied to the column vector W
            G2 = atoms.G2[:, None]
            out = W / G2 / -atoms.Omega
            out[0, :] = 0
    return out


@handle_spin_gracefully
def I(atoms, W):
    '''Backwards transformation from reciprocal space to real-space.

    This operator acts on the options 3, 4, 5, and 6.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    n = np.prod(atoms.s)
    # If W is in the full space do nothing with W
    if len(W) == len(atoms.G2):
        Wfft = W
    else:
        # Fill with zeros if W is in the active space
        if W.ndim == 1:
            Wfft = np.zeros(n, dtype=complex)
        else:
            Wfft = np.zeros((n, atoms.Nstate), dtype=complex)
        Wfft[atoms.active] = W

    if W.ndim == 1:
        Wfft = Wfft.reshape(atoms.s)
        Finv = ifftn(Wfft, workers=THREADS).ravel()
    else:
        # Here we reshape the input like in the 1d case but add an extra dimension in the end,
        # holding the number of states
        Wfft = Wfft.reshape(np.append(atoms.s, atoms.Nstate))
        # Tell the function that the FFT only has to act on the first 3 axes
        Finv = ifftn(Wfft, workers=THREADS, axes=(0, 1, 2)).reshape((n, atoms.Nstate))
    return Finv * n


@handle_spin_gracefully
def J(atoms, W, full=True):
    '''Forward transformation from real-space to reciprocal space.

    This operator acts on the options 1 and 2.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        full (bool): Wether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    '''
    n = np.prod(atoms.s)
    if W.ndim == 1:
        Wfft = W.reshape(atoms.s)
        F = fftn(Wfft, workers=THREADS).ravel()
    else:
        Wfft = W.reshape(np.append(atoms.s, atoms.Nstate))
        F = fftn(Wfft, workers=THREADS, axes=(0, 1, 2)).reshape((n, atoms.Nstate))

    # There is no way to know if J has to transform to the full or the active space
    # but normally it transforms to the full space
    if not full:
        F = F[atoms.active]
    return F / n


# No spin handle is needed since J already has one
def Idag(atoms, W, full=False):
    '''Conjugated backwards transformation from real-space to reciprocal space.

    This operator acts on the options 1 and 2.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Keyword Args:
        full (bool): Wether to transform in the full or in the active space.

    Returns:
        ndarray: The operator applied on W.
    '''
    n = np.prod(atoms.s)
    F = J(atoms, W, full)
    return F * n


# No spin handle is needed since I already has one
def Jdag(atoms, W):
    '''Conjugated forward transformation from reciprocal space to real-space.

    This operator acts on the options 3, 4, 5, and 6.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    n = np.prod(atoms.s)
    Finv = I(atoms, W)
    return Finv / n


@handle_spin_gracefully
def K(atoms, W):
    '''Preconditioning operator.

    This operator acts on the options 3 and 5.
    Reference: Comput. Phys. Commun. 128, 1.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.

    Returns:
        ndarray: The operator applied on W.
    '''
    # G2 is a normal 1d row vector, reshape it so it can be applied to the column vector W
    if len(W) == len(atoms.G2c):
        G2 = atoms.G2c[:, None]
    else:
        G2 = atoms.G2[:, None]
    return W / (1 + G2)


@handle_spin_gracefully
def T(atoms, W, dr):
    '''Translation operator.

    This operator acts on the options 5 and 6.

    Args:
        atoms: Atoms object.
        W (ndarray): Expansion coefficients of unconstrained wave functions in reciprocal space.
        dr (ndarray): Real-space shift.

    Returns:
        ndarray: The operator applied on W.
    '''
    # Do the shift by multiplying a phase factor, given by the shift theorem
    factor = np.exp(-1j * atoms.G[atoms.active] @ dr)
    # factor is a normal 1d row vector, reshape it so it can be applied to the column vector W
    if W.ndim == 2:
        factor = factor[:, None]
    return factor * W



#!/usr/bin/env python3
'''Setup file to make eminus installable.

For a full list of options see the documentation:
https://setuptools.pypa.io/en/latest/references/keywords.html
'''
from setuptools import find_packages, setup

with open('eminus/version.py', 'r') as fh:
    version = {}
    exec(fh.read(), version)

with open('README.md', 'r') as readme, open('CHANGELOG.md', 'r') as changelog:
    long_description = readme.read() + '\n\n' + changelog.read()

extras = {
    'fods': [
        'pyflosic2>=2.0.0rc0'  # PyCOM FOD guessing method
    ],
    'libxc': [
        'pyscf>=1.5.1'  # LibXC interface via PySCF
    ],
    'viewer': [
        'nglview>=2.6.5',  # Molecule and isosurface viewer
        'plotly>=4'        # Grid visualization
    ]
}
extras['all'] = [dep for values in extras.values() for dep in values]
extras['dev'] = [
    'flake8>=3.7',               # Style guide checker
    'flake8-docstrings>=1.4',    # Docstring style guide extension
    'flake8-import-order>=0.9',  # Import statement style guide extension
    'furo>=2022.02.14.1',        # Documentation theme
    'notebook',                  # Run and convert notebooks to HTML
    'pytest>=2.8',               # Test utilities
    'sphinx>=4'                  # Documentation builder
]


setup(
    name='eminus',
    version=version['__version__'],
    description='A plane wave density functional theory code.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Wanja Timm Schulze',
    author_email='wangenau@protonmail.com',
    url='https://wangenau.gitlab.io/eminus',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development'
    ],
    license='APACHE2.0',
    install_requires=[
        'numpy>=1.17',
        'scipy>=1.4'
    ],
    keywords=['ESP'],
    package_data={
        'eminus.pade': ['*-q*']
    },
    zip_safe=False,
    extras_require=extras,
    python_requires='>=3.6',
    project_urls={
        'Bug Tracker': 'https://gitlab.com/wangenau/eminus/-/issues',
        'Changelog': 'https://wangenau.gitlab.io/eminus/changelog.html',
        'Source': 'https://gitlab.com/wangenau/eminus'
    }
)



#!/usr/bin/env python3
'''Fermi-orbital descriptor generation.'''
import pathlib

import numpy as np
from scipy.linalg import norm

from ..data import SYMBOL2NUMBER
from ..logger import log
from ..units import ang2bohr, bohr2ang


def get_fods(object, basis='pc-0', loc='FB', clean=True, elec_symbols=None):
    '''Generate FOD positions using the PyCOM method.

    Reference: J. Comput. Chem. 40, 2843.

    Args:
        object: Atoms or SCF object.

    Keyword Args:
        basis (str): Basis set for the DFT calculation.
        loc (str): Localization method (case insensitive).
        clean (bool): Remove log files.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        ndarray: FOD positions.
    '''
    try:
        from pyflosic2.atoms.atoms import Atoms
        from pyflosic2.guess.pycom import pycom
        from pyflosic2.parameters.flosic_parameters import parameters
        from pyscf.gto import M  # PySCF is a dependency of PyFLOSIC2
        from pyscf.scf import UKS, RKS
    except ImportError:
        log.exception('Necessary dependencies not found. To use this module, '
                      'install them with "pip install eminus[fods]".\n\n')
        raise

    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object
    loc = loc.upper()

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to calculate helium in the spin-'
                        'polarized case.')

    # Convert to Angstrom for PySCF
    X = bohr2ang(atoms.X)
    # Build the PySCF input format
    atom_pyscf = list(zip(atoms.atom, X))

    # Spin in PySCF is the difference of up and down electrons
    if atoms.Nspin == 2:
        spin = int(np.sum(atoms.f[0] - atoms.f[1]))
    else:
        spin = np.sum(atoms.Z) % 2

    # Do the PySCF DFT calculation
    mol = M(atom=atom_pyscf, basis=basis, spin=spin)
    if atoms.Nspin == 2:
        mf = UKS(mol=mol)
    else:
        mf = RKS(mol=mol)
    mf.verbose = 0
    mf.kernel()

    # Add some FODs to the positions, otherwise the method will not work
    extra_up = np.zeros((len(np.nonzero(atoms.f[0])[0]), 3))
    atom_pyflosic = atoms.atom + [elec_symbols[0]] * len(extra_up)
    X_pyflosic = np.vstack((X, extra_up))
    if atoms.Nspin == 2:
        extra_dn = np.zeros((len(np.nonzero(atoms.f[1])[0]), 3))
        atom_pyflosic += [elec_symbols[1]] * len(extra_dn)
        X_pyflosic = np.vstack((X_pyflosic, extra_dn))

    # Do the PyCOM FOD generation
    atoms_pyflosic = Atoms(atom_pyflosic, X_pyflosic, elec_symbols=elec_symbols, spin=spin)
    if atoms.Nspin == 2:
        p = parameters(mode='unrestricted')
    else:
        p = parameters(mode='restricted')
    p.init_atoms(atoms_pyflosic)
    p.basis = basis
    p.pycom_loc = loc
    pc = pycom(mf=mf, p=p)
    pc.get_guess()

    fod1 = ang2bohr(pc.p.fod1.positions)
    fod2 = ang2bohr(pc.p.fod2.positions)
    fods = [np.asarray(fod1), np.asarray(fod2)]

    if clean:
        pathlib.Path(p.log_name).unlink()
        pathlib.Path(f'{loc}_GUESS_COM.xyz').unlink()
    return fods


def split_fods(atom, X, elec_symbols=None):
    '''Split atom and FOD coordinates.

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        tuple[list, ndarray, list]: Atom types, the respective coordinates, and FOD positions.
    '''
    if elec_symbols is None:
        elec_symbols = ['X', 'He']

    X_fod_up = []
    X_fod_dn = []
    # Iterate in reverted order, because we may delete elements
    for ia in range(len(X) - 1, -1, -1):
        if atom[ia] in elec_symbols:
            if atom[ia] in elec_symbols[0]:
                X_fod_up.append(X[ia])
            if atom[ia] in elec_symbols[1]:
                X_fod_dn.append(X[ia])
            X = np.delete(X, ia, axis=0)
            del atom[ia]

    X_fod = [np.asarray(X_fod_up), np.asarray(X_fod_dn)]
    return atom, X, X_fod


def remove_core_fods(object, fods):
    '''Remove core FODs from a set of FOD coordinates.

    Args:
        object: Atoms or SCF object.
        fods (ndarray): FOD positions.

    Returns:
        ndarray: Valence FOD positions.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    # If the number of valence electrons is the same as the number of FODs, do nothing
    if atoms.Nspin == 1 and len(fods[0]) == np.sum(atoms.f[0]):
        return fods
    if atoms.Nspin == 2 and len(fods[0]) == np.sum(atoms.f[0]) \
            and len(fods[1]) == np.sum(atoms.f[1]):
        return fods

    for spin in range(atoms.Nspin):
        for ia in range(atoms.Natoms):
            n_core = SYMBOL2NUMBER[atoms.atom[ia]] - atoms.Z[ia]
            # In the spin-paired case two electrons are one state
            # Since only core states are removed in pseudopotentials this value is divisible by 2
            # +1 to account for uneven amount of core FODs (e.g., for hydrogen)
            n_core = (n_core + 1) // 2
            dist = norm(fods[spin] - atoms.X[ia], axis=1)
            idx = np.argsort(dist)
            # Remove core FODs with the smallest distance to the core
            fods[spin] = np.delete(fods[spin], idx[:n_core], axis=0)
    return fods



#!/usr/bin/env python3
'''Import and export functionalities.'''
import inspect
import pathlib
import pickle
import textwrap
import time

import numpy as np

from .data import NUMBER2SYMBOL, SYMBOL2NUMBER
from .logger import log
from .units import ang2bohr, bohr2ang
from .version import __version__


def read(*args, **kwargs):
    '''Unified file reader function.'''
    if args[0].endswith('.xyz'):
        return read_xyz(*args, **kwargs)
    elif args[0].endswith('.cube'):
        return read_cube(*args, **kwargs)
    else:
        log.error('No viable file ending found.')


def write(*args, **kwargs):
    '''Unified file writer function.'''
    if args[1].endswith('.xyz'):
        return write_xyz(*args, **kwargs)
    elif args[1].endswith('.cube'):
        return write_cube(*args, **kwargs)
    elif args[1].endswith('.pdb'):
        return write_pdb(*args, **kwargs)
    else:
        log.error('No viable file ending found.')


def read_xyz(filename):
    '''Load atom species and positions from xyz files.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        filename (str): xyz input file path/name.

    Returns:
        tuple[list, ndarray]: Atom species and positions.
    '''
    if not filename.endswith('.xyz'):
        filename += '.xyz'

    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first line contains the number of atoms
        Natoms = int(lines[0].strip())

        # The second line can contain a comment, print it if available
        comment = lines[1].strip()
        log.info(f'XYZ file comment: "{comment}"')

        atom = []
        X = []
        # Following lines contain atom positions with the format: Atom x-pos y-pos z-pos
        for line in lines[2:2 + Natoms]:
            line_split = line.strip().split()
            atom.append(line_split[0])
            X.append(np.float_(line_split[1:4]))

    # xyz files are in Angstrom, so convert to Bohr
    X = ang2bohr(np.asarray(X))
    return atom, X


def write_xyz(object, filename, fods=None, elec_symbols=None):
    '''Generate xyz files from atoms objects.

    File format definition: https://openbabel.org/wiki/XYZ_%28format%29

    Args:
        object: Atoms or SCF object.
        filename (str): xyz output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.xyz'):
        filename += '.xyz'

    # Convert the coordinates from atomic units to Angstrom
    X = bohr2ang(atoms.X)
    if fods is not None:
        fods = [bohr2ang(i) for i in fods]

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    with open(filename, 'w') as fp:
        # The first line contains the number of atoms.
        # If we add FOD coordinates, add them to the count.
        if fods is None:
            fp.write(f'{atoms.Natoms}\n')
        else:
            fp.write(f'{atoms.Natoms + len(fods[0]) + len(fods[1])}\n')
        # The second line can contains a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n')
        for ia in range(atoms.Natoms):
            fp.write(f'{atoms.atom[ia]:<2s}  {X[ia, 0]: .6f}  {X[ia, 1]: .6f}  {X[ia, 2]: .6f}\n')
        # Add FOD coordinates if desired. The atom symbol will default to X (no atom type).
        if fods is not None:
            for ie in fods[0]:
                fp.write(f'{elec_symbols[0]:<2s}  {ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
            for ie in fods[1]:
                fp.write(f'{elec_symbols[1]:<2s}  {ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
    return


def read_cube(filename):
    '''Load atom and cell data from cube files.

    There is no standard for cube files. The following format has been used.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        filename (str): cube input file path/name.

    Returns:
        tuple[list, ndarray, float, ndarray, int]: Species, positions, charges, cell size, sampling.
    '''
    if not filename.endswith('.cube'):
        filename += '.cube'

    # Atomic units and a cuboidal cell that starts at (0,0,0) are assumed.
    with open(filename, 'r') as fh:
        lines = fh.readlines()

        # The first and second line can contain comments, print them if available
        comment = f'{lines[0].strip()}\n{lines[1].strip()}'
        log.info(f'XYZ file comment: "{comment}"')

        # Line 4 to 6 contain the sampling per axis, and the cell basis vectors with length a/s
        # A cuboidal cell is assumed, so only use the diagonal entries
        s = np.empty(3, dtype=int)
        a = np.empty(3)
        for i, line in enumerate(lines[3:6]):
            line_split = line.strip().split()
            s[i] = line_split[0]
            a[i] = s[i] * np.float_(line_split[i + 1])

        atom = []
        X = []
        Z = []
        # Following lines contain atom positions with the format: atom-id charge x-pos y-pos z-pos
        for line in lines[6:]:
            line_split = line.strip().split()
            # If the first value is not a (positive) integer, we have reached the field data
            if not line_split[0].isdigit():
                break
            atom.append(NUMBER2SYMBOL[int(line_split[0])])
            Z.append(line_split[1])
            X.append(np.float_(line_split[2:5]))

    X = np.asarray(X)
    return atom, X, Z, a, s


def write_cube(object, filename, field, fods=None, elec_symbols=None):
    '''Generate cube files from given field data.

    There is no standard for cube files. The following format has been used to work with VESTA.
    File format definition: https://h5cube-spec.readthedocs.io/en/latest/cubeformat.html

    Args:
        object: Atoms or SCF object.
        filename (str): xyz output file path/name.
        field (ndarray): Real-space field data.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    # Atomic units are assumed, so there is no need for conversion.
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.cube'):
        filename += '.cube'

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    # Our field data has been created in a different order than needed for cube files
    # (triple loop over z,y,x instead of x,y,z), so rearrange it with some index magic.
    idx = []
    for Nx in range(atoms.s[0]):
        for Ny in range(atoms.s[1]):
            for Nz in range(atoms.s[2]):
                idx.append(Nx + Ny * atoms.s[0] + Nz * atoms.s[0] * atoms.s[1])
    idx = np.asarray(idx)

    # Make sure we have real valued data in the correct order
    field = np.real(field[idx])

    with open(filename, 'w') as fp:
        # The first two lines have to be a comment.
        # Print information about the file and program, and the file creation time.
        fp.write(f'File generated with eminus {__version__} on {time.ctime()}\n\n')
        # Number of atoms (int), and origin of the coordinate system (float)
        # The origin is normally at 0,0,0 but we could move our box, so take the minimum
        if fods is None:
            fp.write(f'{atoms.Natoms}  ')
        else:
            fp.write(f'{atoms.Natoms + len(fods[0]) + len(fods[1])}  ')
        fp.write('0.0  0.0  0.0\n')
        # Number of points per axis (int), and vector defining the axis (float)
        # We only have a cuboidal box, so each vector only has one non-zero component
        fp.write(f'{atoms.s[0]}  {atoms.a[0] / atoms.s[0]:.6f}  0.0  0.0\n'
                 f'{atoms.s[1]}  0.0  {atoms.a[1] / atoms.s[1]:.6f}  0.0\n'
                 f'{atoms.s[2]}  0.0  0.0  {atoms.a[2] / atoms.s[2]:.6f}\n')
        # Atomic number (int), atomic charge (float), and atom position (floats) for every atom
        for ia in range(atoms.Natoms):
            fp.write(f'{SYMBOL2NUMBER[atoms.atom[ia]]}  {atoms.Z[ia]:.3f}  '
                     f'{atoms.X[ia, 0]: .6f}  {atoms.X[ia, 1]: .6f}  {atoms.X[ia, 2]: .6f}\n')
        if fods is not None:
            for ie in fods[0]:
                fp.write(f'{SYMBOL2NUMBER[elec_symbols[0]]}  0.000  '
                         f'{ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
            for ie in fods[1]:
                fp.write(f'{SYMBOL2NUMBER[elec_symbols[1]]}  0.000  '
                         f'{ie[0]: .6f}  {ie[1]: .6f}  {ie[2]: .6f}\n')
        # Field data (float) with scientific formatting
        # We have s[0]*s[1] chunks values with a length of s[2]
        for i in range(atoms.s[0] * atoms.s[1]):
            # Print every round of values, so we can add empty lines between them
            data_str = '%+1.6e  ' * atoms.s[2] % tuple(field[i * atoms.s[2]:(i + 1) * atoms.s[2]])
            # Print a maximum of 6 values per row
            # Max width for this formatting is 90, since 6*len('+1.00000e-000  ')=90
            fp.write(f'{textwrap.fill(data_str, width=90)}\n\n')
    return


def save(object, filename):
    '''Save objects in a pickle file.

    This function is for personal use only. Never load a file you haven't saved yourself!

    Args:
        object: Class object.
        filename (str): xyz input file path/name.
    '''
    if not filename.endswith(('.pickle', '.pkl')):
        filename += '.pickle'

    with open(filename, 'wb') as fp:
        pickle.dump(object, fp, pickle.HIGHEST_PROTOCOL)
    return


def load(filename):
    '''Load objects from a pickle file.

    This function is for personal use only. Never load a file you haven't saved yourself!

    Args:
        filename (str): xyz input file path/name.

    Returns:
        Class object.
    '''
    if not filename.endswith(('.pickle', '.pkl')):
        filename += '.pickle'

    with open(filename, 'rb') as fh:
        return pickle.load(fh)


def write_pdb(object, filename, fods=None, elec_symbols=None):
    '''Generate pdb files from atoms objects.

    See `~eminus.io.create_pdb_str` for more information about the pdb file format.

    Args:
        object: Atoms or SCF object.
        filename (str): xyz output file path/name.

    Keyword Args:
        fods (list): FOD coordinates to write.
        elec_symbols (list): Identifier for up and down FODs.

    Returns:
        None.
    '''
    try:
        atoms = object.atoms
    except AttributeError:
        atoms = object

    if not filename.endswith('.pdb'):
        filename += '.pdb'

    if elec_symbols is None:
        elec_symbols = ['X', 'He']
        if 'He' in atoms.atom and atoms.Nspin == 2:
            log.warning('You need to modify "elec_symbols" to write helium with FODs in the spin-'
                        'polarized case.')

    atom = atoms.atom
    X = atoms.X
    if fods is not None:
        atom += [elec_symbols[0]] * len(fods[0]) + [elec_symbols[1]] * len(fods[1])
        if fods[0].shape[0] != 0:
            X = np.vstack((X, fods[0]))
        if fods[1].shape[0] != 0:
            X = np.vstack((X, fods[1]))

    with open(filename, 'w') as fp:
        fp.write(create_pdb_str(atom, X, a=atoms.a))
    return


def create_pdb_str(atom, X, a=None):
    '''Convert atom symbols and positions to the pdb format.

    File format definitions:
        CRYST1: https://www.wwpdb.org/documentation/file-format-content/format33/sect8.html#CRYST1

        ATOM: https://www.wwpdb.org/documentation/file-format-content/format33/sect9.html#ATOM

    Args:
        atom (list): Atom symbols.
        X (ndarray): Atom positions.

    Keyword Args:
        a (float): Cell size.

    Returns:
        str: pdb file format.
    '''
    # Convert Bohr to Angstrom
    X = bohr2ang(X)
    if a is not None:
        a = bohr2ang(a)

    # pdb files have specific numbers of characters for every data with changing justification
    # Write everything explicitly down to not loose track of line lengths
    pdb = ''
    # Create data for a cuboidal cell
    if a is not None:
        pdb += 'CRYST1'                 # 1-6 "CRYST1"
        pdb += f'{a[0]:9.3f}'.rjust(9)  # 7-15 a
        pdb += f'{a[1]:9.3f}'.rjust(9)  # 16-24 b
        pdb += f'{a[2]:9.3f}'.rjust(9)  # 25-33 c
        pdb += f'{90:7.2f}'.rjust(7)    # 34-40 alpha
        pdb += f'{90:7.2f}'.rjust(7)    # 41-47 beta
        pdb += f'{90:7.2f}'.rjust(7)    # 48-54 gamma
        pdb += ' '
        pdb += 'P 1'.ljust(11)          # 56-66 Space group
        # pdb += '1'.rjust(4)           # 67-70 Z value
        pdb += '\n'

    # Create molecule data
    pdb += 'MODEL 1'
    for ia in range(len(atom)):
        pdb += '\nATOM  '                   # 1-6 "ATOM"
        pdb += f'{ia + 1}'.rjust(5)         # 7-11 Atom serial number
        pdb += ' '
        pdb += f'{atom[ia]}'.rjust(4)       # 13-16 Atom name
        pdb += ' '                          # 17 Alternate location indicator
        pdb += 'MOL'                        # 18-20 Residue name
        pdb += ' '
        pdb += ' '                          # 22 Chain identifier
        pdb += '1'.rjust(4)                 # 23-26 Residue sequence number
        pdb += ' '                          # 27 Code for insertions of residues
        pdb += '   '
        pdb += f'{X[ia, 0]:8.3f}'.rjust(8)  # 31-38 X orthogonal coordinate
        pdb += f'{X[ia, 1]:8.3f}'.rjust(8)  # 39-46 Y orthogonal coordinate
        pdb += f'{X[ia, 2]:8.3f}'.rjust(8)  # 47-54 Z orthogonal coordinate
        pdb += f'{1:6.2f}'.rjust(6)         # 55-60 Occupancy
        pdb += f'{0:6.2f}'.rjust(6)         # 61-66 Temperature factor
        pdb += '          '
        pdb += f'{atom[ia]}'.rjust(2)       # 77-78 Element symbol
        # pdb += '  '                       # 79-80 Charge
    return f'{pdb}\nENDMDL'


def read_gth(atom, charge=None, psp_path=None):
    '''Read GTH files for a given atom.

    Reference: Phys. Rev. B 54, 1703.

    Args:
        atom (str): Atom name.

    Keyword Args:
        charge (int): Valence charge.
        psp_path (str): Path to GTH pseudopotential files. Defaults to installation_path/pade/.

    Returns:
        dict: GTH parameters.
    '''
    if psp_path is None:
        file_path = pathlib.Path(inspect.getfile(inspect.currentframe())).parent
        psp_path = file_path.joinpath('pade')

    if charge is not None:
        f_psp = psp_path.joinpath(f'{atom}-q{charge}')
    else:
        files = sorted(psp_path.glob(f'{atom}-q*'))
        try:
            f_psp = pathlib.Path(files[0])
        except IndexError:
            log.exception(f'There is no GTH pseudopotential in {psp_path} for "{atom}"')
            raise
        if len(files) > 1:
            log.info(f'Multiple pseudopotentials found for "{atom}". '
                     f'Continue with "{f_psp.name}".')

    psp = {}
    cloc = np.zeros(4)
    rp = np.zeros(4)
    Nproj_l = np.zeros(4, dtype=int)
    h = np.zeros((4, 3, 3))
    try:
        with open(f_psp, 'r') as fh:
            atom = fh.readline()
            # Skip the first line, since we don't need the atom symbol here. If needed, use
            # psp['atom'] = atom.split()[0]  # Atom symbol
            N_all = fh.readline().split()
            psp['Zion'] = sum([int(N) for N in N_all])  # Ionic charge
            loc = fh.readline().split()
            psp['rloc'] = float(loc[0])  # Range of local Gaussian charge distribution
            # Skip the number of local coefficients, since we don't need it. If needed, use
            # psp['n_c_local'] = int(loc[1])  # Number of local coefficients
            for i, val in enumerate(loc[2:]):
                cloc[i] = float(val)
            psp['cloc'] = cloc  # Coefficients for the local part
            lmax = int(fh.readline().split()[0])
            psp['lmax'] = lmax  # Maximal angular momentum in the non-local part
            for i in range(lmax):
                proj = fh.readline().split()
                rp[i], Nproj_l[i] = float(proj[0]), int(proj[1])
                for k, val in enumerate(proj[2:]):
                    h[i, 0, k] = float(val)
                for j in range(1, Nproj_l[i]):
                    proj = fh.readline().split()
                    for k, val in enumerate(proj):
                        h[i, j, j + k] = float(val)
                # Copy upper triangle elements to the lower triangle
                for jtmp in range(3):
                    for ktmp in range(i, 3):
                        h[i, ktmp, jtmp] = h[i, jtmp, ktmp]
            psp['rp'] = rp  # Projector radius for each angular momentum
            psp['Nproj_l'] = Nproj_l  # Number of non-local projectors
            psp['h'] = h  # Projector coupling coefficients per AM channel
    except FileNotFoundError:
        log.exception(f'There is no GTH pseudopotential for "{f_psp.name}"')
        raise
    return psp
